/*
 *
 * Author: Tero Kivinen <kivinen@iki.fi>
 *
 * Copyright (c) 2000 Tero Kivinen <kivinen@iki.fi>
 */
/*
 *        Program: LinuxSaneiCompat
 *        $Source: /u/kivinen/sanei-linux-compat/RCS/linux_sanei_compat.c,v $
 *        $Author: kivinen $
 *
 *        Creation          : 05:17 May  3 2000 kivinen
 *        Last Modification : 12:39 Nov 11 2002 kivinen
 *        Last check in     : $Date: 2002/11/11 10:40:14 $
 *        Revision number   : $Revision: 1.9 $
 *        State             : $State: Exp $
 *        Version           : 1.1054
 *        Edit time         : 497 min
 *
 *        Description       : Linux Scanner compat LKM driver
 *
 *
 *        $Log: linux_sanei_compat.c,v $
 *        Revision 1.9  2002/11/11 10:40:14  kivinen
 *        	Updated to NetBSD-1.6I current.
 *
 *        Revision 1.8  2002/10/08 16:24:24  kivinen
 *        	Added some printfs.
 *
 *        Revision 1.7  2001/08/24 21:23:05  kivinen
 *        	Updated to work with latest netbsd-1.5-current.
 *
 *        Revision 1.6  2001/07/21 13:04:53  kivinen
 *        	Fixed error handler setting if using the new psw_error system.
 *
 *        Revision 1.5  2001/07/21 12:09:10  kivinen
 *        	Ignore file does not exists error when deleting.
 *
 *        Revision 1.4  2001/07/21 12:05:36  kivinen
 *        	Fixed bugs with spl handling (it was not restored if error
 *        	occurred).
 *
 *        Revision 1.3  2001/07/21 10:25:21  kivinen
 *        	Fixed error handling.
 *
 *        Revision 1.2  2001/07/21 09:39:03  kivinen
 *        	Updated version number. Updated to work 1.5 current (i.e with
 *        	new scsi system). Added lsc_hexl_buffer. Added more debugs.
 *        	Fixed lsc_stat to return 0.
 *
 *        Revision 1.1  2000/05/27 14:49:45  kivinen
 *        	Created.
 *
 *        $EndLog$
 */

#define KERNEL
#define _KERNEL
#define _LKM

#include "lscconfig.h"
#include <sys/systm.h>
#include <sys/param.h>
#include <sys/conf.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <sys/proc.h>
#include <sys/uio.h>
#include <sys/kernel.h>
#include <sys/vnode.h>
#include <sys/namei.h>
#include <sys/malloc.h>
#include <sys/mount.h>
#include <sys/exec.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/lkm.h>
#include <sys/fcntl.h>
#include <sys/poll.h>
#include <sys/select.h>
#include <sys/filio.h>
#include <sys/ioctl.h>
#include <sys/device.h>
#include <dev/scsipi/scsipi_all.h>
#include <dev/scsipi/scsiconf.h>

extern int lkmenodev();
extern struct cfdriver scsibus_cd;
extern int nchrdev;

/* A cdevsw structure for the character device. */
extern struct cdevsw lsc_chardev_cdevsw;

#if !defined(VOP_LEASE) && defined(LEASE_CHECK)
#define VOP_LEASE       LEASE_CHECK
#endif

/* Print probe debug messages */
#undef LSC_PROBE_DEBUG
/* Debug scsi commands */
#undef LSC_SG_SCSI_DEBUG
/* Max hex output for each command */
#define LSC_SG_SCSI_DEBUG_MAX_LEN 64
/* Debug scsi commands, but do not debug read commands */
#define LSC_SG_SCSI_DEBUG_NO_READS
/* Debug device reading and writing to/from sg device */
#undef LSC_SG_DEVICE_DEBUG
/* Debug scsi device list device */
#undef LSC_SCSI_DEVICE_DEBUG
/* Debug sg_big_buf device */
#undef LSC_SG_BIG_BUF_DEVICE_DEBUG
/* Generic debug */
#undef LSC_GENERIC_DEBUG
/* Error processing debugging */
#undef LSC_SCSI_ERROR_DEBUG

#if 0
#define LSC_PROBE_DEBUG
#define LSC_SG_SCSI_DEBUG
#define LSC_GENERIC_DEBUG
#endif

#define LSC_LKM_VERSION 	"Linux Sanei Compat V1.2"
#define LSC_LKM_NAME		"LSC"
#define LSC_SG_DEVICE		"/emul/linux/dev/sg"
#define LSC_SCSI_DEVICE		"/emul/linux/proc/scsi/scsi"
#define LSC_SG_BIG_BUF_DEVICE	"/emul/linux/proc/sys/kernel/sg-big-buff"

#define LSC_SG_TIMEOUT 		30000

#define LSC_SCSI_DEVICE_MINOR		0
#define LSC_SG_BIG_BUF_DEVICE_MINOR	1
#define LSC_SG_DEVICE_MINOR		8
#define LSC_MAX_SCANNERS		8

int lsc_number_of_sg_devices = 0;

#define LSC_SG_BIG_BUF_SIZE		130560

int lsc_chardev_is_open = 0;
int lsc_major = 0;

int lsc_scsi_offset = 0;
int lsc_scsi_len = 0;
unsigned char *lsc_scsi_probe = NULL;

int lsc_sg_big_buf_offset = 0;
int lsc_sg_big_buf_len = 0;
unsigned char *lsc_sg_big_buf = NULL;

#ifdef HAVE_SC_MAXTARGET
struct scsipi_link *lsc_link[LSC_MAX_SCANNERS];
#define xs_link sc_link
#else
#ifdef HAVE_SC_CHANNEL
struct scsipi_periph *lsc_link[LSC_MAX_SCANNERS];
#define sc_maxtarget sc_channel->chan_ntargets
#define sc_maxlun sc_channel->chan_nluns
#define scsipi_link scsipi_periph
#define sc_link sc_channel->chan_periphs
#define type periph_type
#define xs_link xs_periph

#else /* HAVE_SC_CHANNEL */
#error "Unsupported version of kernel"
#endif /* HAVE_SC_CHANNEL */
#endif /* HAVE_SC_MAXTARGET */

#ifdef HAVE_DEVICE_ERR_HANDLER
/* We have sc_link->device->err_handler. */
#define err_handler device->err_handler
int (*lsc_old_err_handler[LSC_MAX_SCANNERS]) __P((struct scsipi_xfer *));

#else /* HAVE_DEVICE_ERR_HANDLER */
#ifdef HAVE_PERIPH_SWITCH_PSW_ERROR
#define SCSIRET_CONTINUE EJUSTRETURN
const struct scsipi_periphsw *lsc_old_handlers[LSC_MAX_SCANNERS];
struct scsipi_periphsw lsc_current_handlers[LSC_MAX_SCANNERS];

#else /* HAVE_PERIPH_SWITCH_PSW_ERROR */
#error "Unsupported error handling"
#endif /* HAVE_PERIPH_SWITCH_PSW_ERROR */
#endif /* HAVE_DEVICE_ERR_HANDLER */

unsigned char *lsc_sg_buf_in[LSC_MAX_SCANNERS];
unsigned char *lsc_sg_buf_out[LSC_MAX_SCANNERS];
int lsc_sg_buf_offset[LSC_MAX_SCANNERS];
int lsc_sg_buf_len[LSC_MAX_SCANNERS];
typedef enum {
  LSC_SG_CLOSED = 0,
  LSC_SG_READING = 1,		/* Reading command from user */
  LSC_SG_WRITING = 2		/* Writing reply back to user */
} LscSgStatus;
LscSgStatus lsc_sg_status[LSC_MAX_SCANNERS];

int lsc_sg_timeout[LSC_MAX_SCANNERS];

MOD_DEV(LSC_LKM_VERSION, LSC_SG_DEVICE, NULL, -1,
	&lsc_chardev_cdevsw, -1);

const char *lsc_scsi_linux_type[T_NODEVICE + 1] =
{
  "Direct-Access    ",		/* T_DIRECT */
  "Sequential-Access",		/* T_SEQUENTIAL */
  "Printer          ",		/* T_PRINTER */
  "Processor        ",		/* T_PROCESSOR */
  "WORM             ",		/* T_WORM */
  "CD-ROM           ",		/* T_CDROM */
  "Scanner          ",		/* T_SCANNER */
  "Optical Device   ",		/* T_OPTICAL */
  "Medium Changer   ",		/* T_CHANGER */
  "Communications   ",		/* T_COMM */
  "Unknown          ",		/* T_IT8_1 */
  "Unknown          ",		/* T_IT8_2 */
  "Unknown          ",		/* T_STORARRAY */
  "Enclosure        ",		/* T_ENCLOSURE */
  "Unknown          ",		/* 0x0e */
  "Unknown          ",		/* 0x0f */
  "Unknown          ",		/* 0x10 */
  "Unknown          ",		/* 0x11 */
  "Unknown          ",		/* 0x12 */
  "Unknown          ",		/* 0x13 */
  "Unknown          ",		/* 0x14 */
  "Unknown          ",		/* 0x15 */
  "Unknown          ",		/* 0x16 */
  "Unknown          ",		/* 0x17 */
  "Unknown          ",		/* 0x18 */
  "Unknown          ",		/* 0x19 */
  "Unknown          ",		/* 0x1a */
  "Unknown          ",		/* 0x1b */
  "Unknown          ",		/* 0x1c */
  "Unknown          ",		/* 0x1d */
  "Unknown          ",		/* 0x1e */
  "No device        "		/* 0x1f */
};

/* Hexl dump data buffer */
void lsc_hexl_buffer(const unsigned char *buffer, size_t len)
{
  int i, j;
  
  for(i = 0; i < len; i += 16)
    {
      if (i >= LSC_SG_SCSI_DEBUG_MAX_LEN)
	{
	  printf("...\n");
	  break;
	}
      printf("%08x: ", i);
      for(j = 0; j < 16; j++)
	{
	  if (i + j >= len)
	    printf("  ");
	  else
	    printf("%02x", buffer[i + j]);
	  if (j % 2 == 1)
	    printf(" ");
	}
      printf(" ");
      for(j = 0; j < 16; j++)
	{
	  if (i + j >= len)
	    printf(" ");
	  else if (buffer[i + j] >= ' ' &&
		   buffer[i + j] <= '~')
	    printf("%c", buffer[i + j]);
	  else
	    printf(".");
	}
      printf("\n");
    }
}

/* SPL to use to protect internal data. */
int lsc_spl()
{
  return splhigh();
}

/* Removes the device from the file system. This function is used when loading
   and unloading the driver. */
static int lsc_remove(const char *name)
{
  struct nameidata nd;
  int error;

  NDINIT(&nd, DELETE, LOCKPARENT | LOCKLEAF, UIO_SYSSPACE, name, curproc);
  if ((error = namei(&nd)))
    {
      if (error != ENOENT)
	printf("Namei failed for file %s, error = %d\n", name, error);
      return error;
    }
  VOP_LEASE(nd.ni_dvp, curproc, curproc->p_ucred, LEASE_WRITE);
  VOP_LEASE(nd.ni_vp, curproc, curproc->p_ucred, LEASE_WRITE);
  error = VOP_REMOVE(nd.ni_dvp, nd.ni_vp, &nd.ni_cnd);
  if (error)
    printf("Remove failed for file %s, error = %d\n", name, error);
  return error;
}

/* Removes all devices create from the file system. This function is used when
   loading and unloading the driver. */
static int lsc_remove_all(void)
{
  char buffer[256];
  int i;

  for(i = 0; i < 16; i++)
    {
      snprintf(buffer, sizeof(buffer), "%s%d", LSC_SG_DEVICE, i);
      lsc_remove(buffer);
    }
  lsc_remove(LSC_SCSI_DEVICE);
  lsc_remove(LSC_SG_BIG_BUF_DEVICE);
}

/* Create the device to the file system. This function is used when loading the
   driver */
static int lsc_create(const char *name, int minor_number)
{
  struct nameidata nd;
  struct vattr vattr;
  int error = 0;

  NDINIT(&nd, CREATE, LOCKPARENT, UIO_SYSSPACE, name, curproc);
  if ((error = namei(&nd)))
    {
      printf("Namei failed for file %s, error = %d\n", name, error);
      return error;
    }
  if (nd.ni_vp != NULL)
    {
      VOP_ABORTOP(nd.ni_dvp, &nd.ni_cnd);
      if (nd.ni_dvp == nd.ni_vp)
        vrele(nd.ni_dvp);
      else
        vput(nd.ni_dvp);
      vrele(nd.ni_vp);
      printf("File %s already exists\n", name);
      return EEXIST;
    }
  VATTR_NULL(&vattr);
  vattr.va_type = VCHR;
  vattr.va_mode = 0666;
  vattr.va_rdev = makedev(lsc_major, minor_number);
  VOP_LEASE(nd.ni_dvp, curproc, curproc->p_ucred, LEASE_WRITE);
  error = VOP_MKNOD(nd.ni_dvp, &nd.ni_vp, &nd.ni_cnd, &vattr);
  if (error)
    {
      printf("Making of the device failed for file %s, error = %d\n",
	     name, error);
      return error;
    }
  vput(nd.ni_vp);
  
  return 0;
}

/* Create all devices to the file system. This function is used when loading
   the driver */
static int lsc_create_all(void)
{
  char buffer[256];
  int error;
  int i;

  for(i = 0; i < lsc_number_of_sg_devices; i++)
    {
      snprintf(buffer, sizeof(buffer), "%s%d", LSC_SG_DEVICE, i);
      error = lsc_create(buffer, LSC_SG_DEVICE_MINOR + i);
      if (error)
	return error;
    }
  error = lsc_create(LSC_SCSI_DEVICE, LSC_SCSI_DEVICE_MINOR);
  if (error)
    return error;
  error = lsc_create(LSC_SG_BIG_BUF_DEVICE, LSC_SG_BIG_BUF_DEVICE_MINOR);
  if (error)
    return error;
}

/* This function is called after loading the lsc driver into memory. This will
   create the device that the lsc driver uses to communicate with user-mode
   processes. */
static int lsc_load(struct lkm_table *lkmtp, int cmd)
{
  struct lkm_dev *args = lkmtp->private.lkm_dev;
  int error = 0;
  int i;

  if (lkmexists(lkmtp))
    return EEXIST;

  lsc_sg_big_buf = malloc(16, M_TEMP, M_NOWAIT);
  if (lsc_sg_big_buf == NULL)
    return ENOMEM;
  snprintf(lsc_sg_big_buf, 16, "%u\n", LSC_SG_BIG_BUF_SIZE);

  error = devsw_attach(args->lkm_devname,
		       args->lkm_bdev, &args->lkm_bdevmaj,
		       args->lkm_cdev, &args->lkm_cdevmaj);
  if (error != 0)
    return (error);
  lsc_major = args->lkm_cdevmaj;
  devsw_detach(args->lkm_bdev, args->lkm_cdev);

  args->lkm_offset = i;   /* slot in cdevsw[] */

  lsc_number_of_sg_devices = 0;
  
#ifdef LSC_PROBE_DEBUG
  printf("%s: scsibus_cd:data\n", LSC_LKM_NAME);
  lsc_hexl_buffer((void *) &scsibus_cd, sizeof(scsibus_cd));
  printf("%s: There are %d scsi busses, busname = %s\n", LSC_LKM_NAME, 
	     scsibus_cd.cd_ndevs, scsibus_cd.cd_name);
#endif /* LSC_PROBE_DEBUG */
  for(i = 0; i < scsibus_cd.cd_ndevs; i++)
    {
      struct scsibus_softc *scsi;
      struct scsipi_link *link;
      int target, lun;

      scsi = scsibus_cd.cd_devs[i];
      if (scsi == NULL)
	continue;

#ifdef HAVE_SC_CHANNEL
      if (scsi->sc_channel == NULL)
	continue;
#endif /* HAVE_SC_CHANNEL */
      
      if (scsi->sc_maxtarget > 16 || scsi->sc_maxlun > 16)
	{
	  printf("%s: Maxtarget (%d) or maxlun (%d) too large, perhaps the "
		 "kernel is newer than the lsc binary\n", LSC_LKM_NAME,
		 scsi->sc_maxtarget, scsi->sc_maxlun);
	  free(lsc_sg_big_buf, M_TEMP);
	  return ENODEV;
	}
      for(target = 0; target < scsi->sc_maxtarget; target++)
	for(lun = 0; lun < scsi->sc_maxlun; lun++)
	  {	
	    struct scsipi_link *link;
	    struct scsipi_inquiry_data inqbuf;
	    int error;

	    link = scsipi_lookup_periph(scsi->sc_channel, target, lun);
	    if (link == NULL)
	      continue;
#ifdef HAVE_SC_MAXTARGET
	    if (link->type != BUS_SCSI)
	      {
#ifdef LSC_PROBE_DEBUG
		printf("Target %d lun %d type not BUS_SCSI (%d) = %d\n",
		       target, lun, BUS_SCSI, link->type);
#endif /* LSC_PROBE_DEBUG */
		continue;
	      }
#endif /* HAVE_SC_MAXTARGET */
	    memset(&inqbuf, 0, sizeof(inqbuf));
	    error = scsipi_inquire(link, &inqbuf, 0);
	    if (error)
	      {
#ifdef LSC_PROBE_DEBUG
		printf("%s: Bus %d, target %d, lun %d scsipi_inquire returned error %d\n",
		       LSC_LKM_NAME, i, target, lun, error);
#endif /* LSC_PROBE_DEBUG */
		error = scsipi_inquire(link, &inqbuf, XS_CTL_RESET);
		if (error)
		  {
#ifdef LSC_PROBE_DEBUG
		    printf("%s: Bus %d, target %d, lun %d scsipi_inquire returned error with XS_CTL_RESET %d\n",
			   LSC_LKM_NAME, i, target, lun, error);
#endif /* LSC_PROBE_DEBUG */
		    continue;
		  }
	      }
#ifdef LSC_PROBE_DEBUG
	    printf("%s: Bus %d, target %d, lun %d scsipi_inquire says device is (%d) %s\n",
		   LSC_LKM_NAME, i, target, lun,
		   inqbuf.device & SID_TYPE, 
		   lsc_scsi_linux_type[(unsigned long) inqbuf.device &
				      SID_TYPE]);
#endif /* LSC_PROBE_DEBUG */
	    if ((inqbuf.device & SID_TYPE) == T_SCANNER)
	      {
		if (lsc_number_of_sg_devices < LSC_MAX_SCANNERS)
		  {
		    lsc_link[lsc_number_of_sg_devices] = link;
		    lsc_sg_status[lsc_number_of_sg_devices] = LSC_SG_CLOSED;
		    lsc_sg_timeout[lsc_number_of_sg_devices] = LSC_SG_TIMEOUT;
		  }
		lsc_number_of_sg_devices++;
	      }
	  }
    }
  printf("%s loaded, major dev %d, found %d scanners\n",
         LSC_LKM_VERSION, lsc_major, lsc_number_of_sg_devices);

  if (lsc_number_of_sg_devices > LSC_MAX_SCANNERS)
    {
      printf("%s: Number of scanners too large, recompile with "
	     "larger LSC_MAX_SCANNERS value, current = %d\n", LSC_LKM_NAME,
	     LSC_MAX_SCANNERS);
      lsc_number_of_sg_devices = LSC_MAX_SCANNERS;
    }

  /* Remove any existing devices with the name used by the driver. */
  lsc_remove_all();

  /* Create all devices */
  error = lsc_create_all();
  if (error)
    {
      lsc_remove_all();
      return error;
    }

  return 0;
}

/* This function is called just before unloading the driver from memory. */
static int lsc_unload(struct lkm_table *lkmtp, int cmd)
{
  struct lkm_dev *args = lkmtp->private.lkm_dev;

  /* If the driver is currently open as a device, refuse to unload it. */
  if (lsc_chardev_is_open)
    return EBUSY;

  free(lsc_sg_big_buf, M_TEMP);

  lsc_remove_all();

  printf("%s: unloaded from major dev %d\n", LSC_LKM_NAME, lsc_major);

  devsw_detach(args->lkm_bdev, args->lkm_cdev);
  args->lkm_bdevmaj = -1;
  args->lkm_cdevmaj = -1;
  return 0;
}

/* Status query entry point. */
static int lsc_stat(struct lkm_table *lkmtp, int cmd)
{
  return 0;
}

/* This is the public entry point that is called first when the module
   is loaded. */
int lsc(lkmtp, cmd, ver)
struct lkm_table *lkmtp;
int cmd, ver;
{
  DISPATCH(lkmtp, cmd, ver, lsc_load, lsc_unload, lsc_stat);
}

#define ENTRY_FUNC(name) \
int name(lkmtp, cmd, ver) \
struct lkm_table *lkmtp; \
int cmd, ver; \
{ \
  return lsc(lkmtp, cmd, ver); \
}

/* Old NetBSD */
ENTRY_FUNC(xxxinit)

/* New NetBSD */
ENTRY_FUNC(lsc_lkmentry)
     
int lsc_error_handler(struct scsipi_xfer *xs)
{
  int i;

  for(i = 0; i < lsc_number_of_sg_devices; i++)
    {
      if (xs->xs_link == lsc_link[i] && lsc_sg_status != LSC_SG_CLOSED)
	{
	  memcpy(lsc_sg_buf_out[i] + 20, &xs->sense.scsi_sense, 16);
#ifdef LSC_SCSI_ERROR_DEBUG
	  printf("%s: Error in scsi command control = %x, status = %x, " 
		 "retries = %d, reqcnt = %d, timeout = %d, "
		 "cmd(%d), data(%d), "
		 "resid = %d, error = %d\n",
		 LSC_LKM_NAME,
		 xs->xs_control,
		 xs->xs_status,
		 xs->xs_retries,
		 xs->xs_requeuecnt,
		 xs->timeout,
		 xs->cmdlen,
		 xs->datalen, 
		 xs->resid,
		 xs->error);
	  if (xs->cmd != NULL)
	    {
	      printf("%s: cmd\n", LSC_LKM_NAME);
	      lsc_hexl_buffer((unsigned char *) xs->cmd, xs->cmdlen);
	    }
	  if (xs->data != NULL)
	    {
	      printf("%s: Data\n", LSC_LKM_NAME);
	      lsc_hexl_buffer(xs->data, xs->datalen);
	    }
	  printf("%s: Sense\n", LSC_LKM_NAME);
	  lsc_hexl_buffer((unsigned char *) &xs->sense.scsi_sense, 32);
#endif /* LSC_SCSI_ERROR_DEBUG */
	  return SCSIRET_CONTINUE;
	}
    }
  printf("%s: Error, cannot find matching device for the error\n",
	 LSC_LKM_NAME);
  return SCSIRET_CONTINUE;
}
     
/* This is called whenever the device is opened.  This returns
   errno, or 0 on successs. */
int lsc_cdevsw_open(dev_t dev, int flag, int mode, struct proc *p)
{
  struct scsibus_softc *scsi;
  struct scsipi_link *link;
  int target, lun;
  int s, i;

  s = lsc_spl();

#ifdef LSC_GENERIC_DEBUG
  printf("%s: Opening device, minor = %d\n", LSC_LKM_NAME, minor(dev));
#endif /* LSC_GENERIC_DEBUG */
  if (minor(dev) == LSC_SCSI_DEVICE_MINOR)
    {
      int num_scsi_devs;
      size_t len;
      char *ptr;

      if (lsc_scsi_len != 0)
	{
	  splx(s);
	  return EBUSY;
	}
#ifdef LSC_SCSI_DEVICE_DEBUG
      printf("%s: Found %d scsi busses, busname = %s\n", LSC_LKM_NAME, 
	     scsibus_cd.cd_ndevs, scsibus_cd.cd_name);
#endif /* LSC_SCSI_DEVICE_DEBUG */
      num_scsi_devs = 0;
      for(i = 0; i < scsibus_cd.cd_ndevs; i++)
	{
	  scsi = scsibus_cd.cd_devs[i];
	  if (scsi == NULL)
	    continue;
#ifdef HAVE_SC_CHANNEL
	  if (scsi->sc_channel == NULL)
	    continue;
#endif /* HAVE_SC_CHANNEL */
	  for(target = 0; target < scsi->sc_maxtarget; target++)
	    for(lun = 0; lun < scsi->sc_maxlun; lun++)
	      {
		link = scsipi_lookup_periph(scsi->sc_channel, target, lun);
		if (link == NULL)
		  continue;
#ifdef HAVE_SC_MAXTARGET
		if (link->type != BUS_SCSI)
		  continue;
#endif /* HAVE_SC_MAXTARGET */
		num_scsi_devs++;
	      }
	}
#ifdef LSC_SCSI_DEVICE_DEBUG
      printf("%s: Found %d scsi devices\n", LSC_LKM_NAME, num_scsi_devs);
#endif /* LSC_SCSI_DEVICE_DEBUG */
      len = 200 * num_scsi_devs + 50;
      lsc_scsi_probe = malloc(len, M_TEMP, M_NOWAIT);
      if (lsc_scsi_probe == NULL)
	{
	  splx(s);
	  return ENOMEM;
	}
      ptr = lsc_scsi_probe;
      strcpy(ptr, "Attached devices: \n");
      ptr += strlen(ptr);
      for(i = 0; i < scsibus_cd.cd_ndevs; i++)
	{
	  scsi = scsibus_cd.cd_devs[i];
	  if (scsi == NULL)
	    continue;
#ifdef HAVE_SC_CHANNEL
	  if (scsi->sc_channel == NULL)
	    continue;
#endif /* HAVE_SC_CHANNEL */
#ifdef LSC_SCSI_DEVICE_DEBUG
	  printf("%s: Device %s, max targes = %d, max luns = %d\n",
		 LSC_LKM_NAME, 
		 scsi->sc_dev.dv_xname,
		 scsi->sc_maxtarget, scsi->sc_maxlun);
#endif /* LSC_SCSI_DEVICE_DEBUG */
	  for(target = 0; target < scsi->sc_maxtarget; target++)
	    for(lun = 0; lun < scsi->sc_maxlun; lun++)
	      {
		struct scsipi_link *link;
		struct scsipi_inquiry_data inqbuf;
		int error;

		link = scsipi_lookup_periph(scsi->sc_channel, target, lun);
		if (link == NULL)
		  continue;
#ifdef HAVE_SC_MAXTARGET
		if (link->type != BUS_SCSI)
		  continue;
#endif /* HAVE_SC_MAXTARGET */
#ifdef LSC_SCSI_DEVICE_DEBUG
		printf("%s: Found target %d, lun %d\n", LSC_LKM_NAME,
		       target, lun);
#endif /* LSC_SCSI_DEVICE_DEBUG */

		memset(&inqbuf, 0, sizeof(inqbuf));
		error = scsipi_inquire(link, &inqbuf, 0);
		if (error)
		  {
		    printf("%s: Error in scsi identity command: %d\n",
			   LSC_LKM_NAME, error);
		    error = scsipi_inquire(link, &inqbuf, XS_CTL_RESET);
		    if (error)
		      continue;
		  }
		sprintf(ptr, "Host: scsi%d Channel: %02d "
			"Id: %02d Lun: %02d\n",
			i, i, target, lun);
		ptr += strlen(ptr);
		sprintf(ptr, "  Vendor: %-8.8s Model: %-16.16s "
			"Rev: %-4.4s\n",
			inqbuf.vendor, inqbuf.product, inqbuf.revision);
		ptr += strlen(ptr);
		sprintf(ptr,"  Type:   %-17.17s                "
			"ANSI SCSI revision: %02d\n",
			(lsc_scsi_linux_type[(inqbuf.device & SID_TYPE)]),
			(inqbuf.version & SID_ANSII));
		ptr += strlen(ptr);
	      }
	}
      lsc_scsi_len = strlen(lsc_scsi_probe);
      lsc_scsi_offset = 0;
    }
  else if (minor(dev) == LSC_SG_BIG_BUF_DEVICE_MINOR)
    {
      if (lsc_sg_big_buf_len != 0)
	{
	  splx(s);
	  return EBUSY;
	}
      lsc_sg_big_buf_len = strlen(lsc_sg_big_buf);
      lsc_sg_big_buf_offset = 0;
    }
  else if (minor(dev) >= LSC_SG_DEVICE_MINOR &&
	   minor(dev) < LSC_SG_DEVICE_MINOR + lsc_number_of_sg_devices)
    {
      int ind;

      ind = minor(dev) - LSC_SG_DEVICE_MINOR;

      if (lsc_sg_status[ind] != LSC_SG_CLOSED)
	{
	  splx(s);
	  return EBUSY;
	}
#ifdef HAVE_PERIPH_SWITCH_PSW_ERROR
      lsc_old_handlers[ind] = lsc_link[ind]->periph_switch;
      memcpy(&lsc_current_handlers[ind], lsc_link[ind]->periph_switch,
	     sizeof(lsc_current_handlers[ind]));
      lsc_current_handlers[ind].psw_error = lsc_error_handler;
      lsc_link[ind]->periph_switch = &lsc_current_handlers[ind];
#else /* HAVE_PERIPH_SWITCH_PSW_ERROR */
      lsc_old_err_handler[ind] = lsc_link[ind]->err_handler;
      lsc_link[ind]->err_handler = lsc_error_handler;
#endif /* HAVE_PERIPH_SWITCH_PSW_ERROR */
      lsc_sg_buf_len[ind] = 0;
      lsc_sg_buf_offset[ind] = 0;
      lsc_sg_buf_in[ind] = malloc(LSC_SG_BIG_BUF_SIZE, M_TEMP, M_NOWAIT);
      if (lsc_sg_buf_in[ind] == NULL)
	{
	  splx(s);
	  return ENOMEM;
	}
      lsc_sg_buf_out[ind] = malloc(LSC_SG_BIG_BUF_SIZE, M_TEMP, M_NOWAIT);
      if (lsc_sg_buf_out[ind] == NULL)
	{
	  free(lsc_sg_buf_in[ind], M_TEMP);
	  splx(s);
	  return ENOMEM;
	}
      lsc_sg_status[ind] = LSC_SG_READING;
    }
  else
    {
      printf("%s: Opening unknown device %d\n", LSC_LKM_NAME, minor(dev));
      splx(s);
      return EIO;
    }
  /* Mark that we are open, and otherwise initialize state. */
  lsc_chardev_is_open++;

#ifdef LSC_GENERIC_DEBUG
  printf("%s: Number of open devices = %d\n",
	 LSC_LKM_NAME, lsc_chardev_is_open);
#endif /* LSC_GENERIC_DEBUG */
  splx(s);
  return 0;
}

/* This is called when the character device is closed for the last
   time (since we only one process to have it open at a time, this
   means whenever it is closed). */
int lsc_cdevsw_close(dev_t dev, int flag, int mode, struct proc *p)
{
  int s;
  
  s = lsc_spl();
  
#ifdef LSC_GENERIC_DEBUG
  printf("%s: Closing device, minor = %d, open devices = %d\n",
	 LSC_LKM_NAME, minor(dev), lsc_chardev_is_open);
#endif /* LSC_GENERIC_DEBUG */

  if (minor(dev) == LSC_SCSI_DEVICE_MINOR)
    {
      lsc_scsi_len = 0;
      free(lsc_scsi_probe, M_TEMP);
    }
  else if (minor(dev) == LSC_SG_BIG_BUF_DEVICE_MINOR)
    {
      lsc_sg_big_buf_len = 0;
    }
  else if (minor(dev) >= LSC_SG_DEVICE_MINOR &&
	   minor(dev) < LSC_SG_DEVICE_MINOR + lsc_number_of_sg_devices)
    {
      int ind;

      ind = minor(dev) - LSC_SG_DEVICE_MINOR;

      lsc_sg_status[ind] = LSC_SG_CLOSED;
      free(lsc_sg_buf_in[ind], M_TEMP);
      free(lsc_sg_buf_out[ind], M_TEMP);
#ifdef HAVE_PERIPH_SWITCH_PSW_ERROR
      lsc_link[ind]->periph_switch = lsc_old_handlers[ind];
#else /* HAVE_PERIPH_SWITCH_PSW_ERROR */
      lsc_link[ind]->err_handler = lsc_old_err_handler[ind];
#endif /* HAVE_PERIPH_SWITCH_PSW_ERROR */
    }
  else
    {
      printf("%s: Closing unknown device %d\n", LSC_LKM_NAME, minor(dev));
      splx(s);
      return EIO;
    }
  /* Mark that we are closed. */
  if (lsc_chardev_is_open > 0)
    lsc_chardev_is_open--;
#ifdef LSC_GENERIC_DEBUG
  printf("%s: Number of open devices = %d\n", LSC_LKM_NAME,
	 lsc_chardev_is_open);
#endif /* LSC_GENERIC_DEBUG */

  splx(s);
  
  return 0;
}

/* IOCTL entry point for the device. */
int lsc_cdevsw_ioctl(dev_t dev, u_long cmd, caddr_t data, int flag,
                     struct proc *p)
{
#ifdef LSC_GENERIC_DEBUG
  printf("%s: Ioctl for dev %d, cmd = 0x%x\n", LSC_LKM_NAME, minor(dev), cmd);
#endif /* LSC_GENERIC_DEBUG */
  return ENOTTY;
}
  
/* Read entry point. */
int lsc_cdevsw_read(dev_t dev, struct uio *uio, int ioflag)
{
  int error = 0, s;
  size_t len;
  
  s = lsc_spl();

#ifdef LSC_GENERIC_DEBUG
  printf("%s: Reading %u bytes from device %d\n", LSC_LKM_NAME,
	 uio->uio_resid, minor(dev));
#endif /* LSC_GENERIC_DEBUG */
  if (minor(dev) == LSC_SCSI_DEVICE_MINOR)
    {
      if (lsc_scsi_len == -1)
	goto out;
      len = uio->uio_resid; 
      if (len > lsc_scsi_len)
	len = lsc_scsi_len;
#ifdef LSC_SCSI_DEVICE_DEBUG
      printf("%s: Copying %u bytes from scsi_probe command\n", LSC_LKM_NAME,
	     len);
#endif /* LSC_SCSI_DEVICE_DEBUG */
      error = uiomove(lsc_scsi_probe + lsc_scsi_offset, len, uio);
      if (!error)
	{
	  lsc_scsi_len -= len;
	  lsc_scsi_offset += len;
	}
      if (lsc_scsi_len == 0)
	lsc_scsi_len = -1;
    }
  else if (minor(dev) == LSC_SG_BIG_BUF_DEVICE_MINOR)
    {
      if (lsc_sg_big_buf_len == -1)
	goto out;
      len = uio->uio_resid; 
      if (len > lsc_sg_big_buf_len)
	len = lsc_sg_big_buf_len;
#ifdef LSC_SG_BIG_BUF_DEVICE_DEBUG
      printf("%s: Copying %u bytes from sg_big_buf command\n", LSC_LKM_NAME,
	     len);
#endif /* LSC_SG_BIG_BUF_DEVICE_DEBUG */
      error = uiomove(lsc_sg_big_buf + lsc_sg_big_buf_offset, len, uio);
      if (!error)
	{
	  lsc_sg_big_buf_len -= len;
	  lsc_sg_big_buf_offset += len;
	}
      if (lsc_sg_big_buf_len == 0)
	lsc_sg_big_buf_len = -1;
    }
  else if (minor(dev) >= LSC_SG_DEVICE_MINOR &&
	   minor(dev) < LSC_SG_DEVICE_MINOR + lsc_number_of_sg_devices)
    {
      int ind;

      ind = minor(dev) - LSC_SG_DEVICE_MINOR;

      if (lsc_sg_status[ind] != LSC_SG_WRITING)
	{
	  error = EIO;
	  goto out;
	}
      len = uio->uio_resid; 
      if (len > lsc_sg_buf_len[ind])
	len = lsc_sg_buf_len[ind];
#ifdef LSC_SG_DEVICE_DEBUG
      printf("%s: Copying %u bytes from sg buffer\n", LSC_LKM_NAME, len);
#endif /* LSC_SG_DEVICE_DEBUG */
      error = uiomove(lsc_sg_buf_out[ind] + lsc_sg_buf_offset[ind], len, uio);
      if (!error)
	{
	  lsc_sg_buf_len[ind] -= len;
	  lsc_sg_buf_offset[ind] += len;
	}
      if (lsc_sg_buf_len[ind] == 0)
	{
	  lsc_sg_status[ind] = LSC_SG_READING;
	  lsc_sg_buf_offset[ind] = 0;
	}
    }
  else
    {
      printf("%s: Reading from unknown device %d\n", LSC_LKM_NAME, minor(dev));
      error = EIO;
    }
 out:
  splx(s);
  return error;
}

/* This is the write entry point for the device.  The written data contains
   one or more packets to send out (to the network or to protocols).
   Packets must always be sent with a single write.  Sending multiple packets
   with a single write is allowed. */
int lsc_cdevsw_write(dev_t dev, struct uio *uio, int ioflag)
{
  int error = 0, s;
  size_t len, i, j;

  s = lsc_spl();

#ifdef LSC_GENERIC_DEBUG
  printf("%s: Writing %u bytes to device %d\n", LSC_LKM_NAME,
	 uio->uio_resid, minor(dev));
#endif /* LSC_GENERIC_DEBUG */

  if (minor(dev) >= LSC_SG_DEVICE_MINOR &&
      minor(dev) < LSC_SG_DEVICE_MINOR + lsc_number_of_sg_devices)
    {
      int ind;

      ind = minor(dev) - LSC_SG_DEVICE_MINOR;

      if (lsc_sg_status[ind] != LSC_SG_READING)
	{
	  error = EIO;
	  goto out;
	}
      len = uio->uio_resid;
      /* Read the pack_len field */
      if (lsc_sg_buf_offset[ind] <= 4)
	{
	  i = 4;
	  if (len < i)
	    i = len;
#ifdef LSC_SG_DEVICE_DEBUG
	  printf("%s: Reading length of the request\n", LSC_LKM_NAME);
#endif /* LSC_SG_DEVICE_DEBUG */
	  error = uiomove(lsc_sg_buf_in[ind] + lsc_sg_buf_offset[ind], i, uio);
	  if (error)
	    goto out;
	  lsc_sg_buf_offset[ind] += i;
	  len -= i;
	  if (lsc_sg_buf_offset[ind] < 4)
	    goto out;
	  lsc_sg_buf_len[ind] = *(int *) lsc_sg_buf_in[ind];
#ifdef LSC_SG_DEVICE_DEBUG
	  printf("%s: Length of the request is %u bytes\n", LSC_LKM_NAME, len);
#endif /* LSC_SG_DEVICE_DEBUG */
	  if (lsc_sg_buf_len[ind] < 36)
	    {
	      error = EIO;
	      goto out;
	    }
	  lsc_sg_buf_len[ind] -= 4;
	}
      /* Read rest of the header and the command */
      if (len > lsc_sg_buf_len[ind])
	len = lsc_sg_buf_len[ind];
#ifdef LSC_SG_DEVICE_DEBUG
      printf("%s: Copying %u bytes to sg buffer\n", LSC_LKM_NAME, len);
#endif /* LSC_SG_DEVICE_DEBUG */
      error = uiomove(lsc_sg_buf_in[ind] + lsc_sg_buf_offset[ind], len, uio);
      if (!error)
	{
	  lsc_sg_buf_len[ind] -= len;
	  lsc_sg_buf_offset[ind] += len;
	}
      if (lsc_sg_buf_len[ind] == 0)
	{
	  /* Everything read */
	  size_t cmd_len, reply_len;

#if 0
	  printf("%s: Input buffer:\n", LSC_LKM_NAME);
	  lsc_hexl_buffer(lsc_sg_buf_in[ind], 36);
#endif

	  lsc_sg_status[ind] = LSC_SG_WRITING;
	  /* Return len */
	  lsc_sg_buf_len[ind] = *(int *) (lsc_sg_buf_in[ind] + 4);
	  lsc_sg_buf_offset[ind] = 0;
	  lsc_sg_buf_len[ind] &= 0x7fffffff;
	  if (lsc_sg_buf_len[ind] < 0 ||
	      lsc_sg_buf_len[ind] > LSC_SG_BIG_BUF_SIZE)
	    {
	      lsc_sg_buf_len[ind] = LSC_SG_BIG_BUF_SIZE;
	    }
	  reply_len = lsc_sg_buf_len[ind] - 36;
	  cmd_len = (*(int *) (lsc_sg_buf_in[ind])) - 36;
#ifdef LSC_SG_SCSI_DEBUG
#ifdef LSC_SG_SCSI_DEBUG_NO_READS
	  if (lsc_sg_buf_in[ind][36] == 0x28)
	    {
	      printf(".");
	    }
	  else
#endif /* LSC_SG_SCSI_DEBUG_NO_READS */
	    {
	      printf("%s: Issuing command of length %u, "
		     "with reply of length %u\n",
		     LSC_LKM_NAME, cmd_len, reply_len);
	      printf("%s: Command:\n", LSC_LKM_NAME);
	      lsc_hexl_buffer(lsc_sg_buf_in[ind] + 36, cmd_len);
	    }
#endif /* LSC_SG_SCSI_DEBUG */
	  memset(lsc_sg_buf_out[ind] + 20, 0, 16);
	  /* Check the direction, if reply_len == 0 then it is to the scsi
	     device, meaning we need to find out the command and data
	     lengths */ 
	  if (reply_len == 0)
	    {
	      int data_len;

	      data_len = cmd_len;

	      switch (lsc_sg_buf_in[ind][36] >> 5)
		{
		case 0: cmd_len = 6; break;
		case 1: cmd_len = 10; break;
		case 2: cmd_len = 10; break;
		case 3: cmd_len = 0;
		  printf("%s: Warning reserved scsi command tag: 0x%02x\n", 
			 LSC_LKM_NAME, lsc_sg_buf_in[ind][36]);
		  break;
		case 4: cmd_len = 0;
		  printf("%s: Warning reserved scsi command tag: 0x%02x\n", 
			 LSC_LKM_NAME, lsc_sg_buf_in[ind][36]);
		  break;
		case 5: cmd_len = 12; break;
		case 6: cmd_len = 0;
		  printf("%s: Warning vendor scsi command tag: 0x%02x\n",
			 LSC_LKM_NAME, lsc_sg_buf_in[ind][36]);
		  break;
		case 7: cmd_len = 0;
		  printf("%s: Warning vendor scsi command tag: 0x%02x\n",
			 LSC_LKM_NAME, lsc_sg_buf_in[ind][36]);
		  break;
		}
	      if (cmd_len == 0 &&
		  (*(int *) (lsc_sg_buf_in[ind] + 16)) & 0x80000000)
		{
		  cmd_len = 12;
		}
	      if (cmd_len == 0)
		{
		  if (data_len < 10)
		    {
		      cmd_len = 6;
		      if (data_len == 6)
			{
			  reply_len = lsc_sg_buf_in[ind][36 + 5];
			}
		    }
		  else if (data_len < 256)
		    {
		      if (lsc_sg_buf_in[ind][36 + 5] + 6 == data_len)
			cmd_len = 6;
		      else if(lsc_sg_buf_in[ind][36 + 8] + 10 == data_len)
			cmd_len = 10;
		      else if (lsc_sg_buf_in[ind][36 + 9] + 12 == data_len)
			cmd_len = 12;
		      else
			{
			  if (data_len == 12)
			    {
			      cmd_len = 12;
			      reply_len =
				((unsigned long) lsc_sg_buf_in[ind][36 + 8] << 8) +
				lsc_sg_buf_in[ind][36 + 9];
			    }
			  else
			    {
			      cmd_len = 10;
			      
			      if (data_len == 10)
				{
				  reply_len =
				    ((unsigned long) lsc_sg_buf_in[ind][36 + 7] << 8) +
				    lsc_sg_buf_in[ind][36 + 8];
				}
			    }
			}
		    }
		  else if (data_len < 65536)
		    {
		      if (((unsigned long) lsc_sg_buf_in[ind][36 + 4] << 8) + 
			  lsc_sg_buf_in[ind][36 + 5] + 6 == data_len)
			cmd_len = 6;
		      else if(((unsigned long) lsc_sg_buf_in[ind][36+7] << 8) +
			      lsc_sg_buf_in[ind][36 + 8] + 10 == data_len)
			cmd_len = 10;
		      else if (((unsigned long) lsc_sg_buf_in[ind][36+8]<<8) +
			       lsc_sg_buf_in[ind][36 + 9] + 12 == data_len)
			cmd_len = 12;
		      else
			cmd_len = 10;
		    }
		  else 
		    {
		      if (((unsigned long) lsc_sg_buf_in[ind][36 + 3] << 16) + 
			  ((unsigned long) lsc_sg_buf_in[ind][36 + 4] << 8) + 
			  lsc_sg_buf_in[ind][36 + 5] + 6 == data_len)
			cmd_len = 6;
		      else if(((unsigned long) lsc_sg_buf_in[ind][36+4]<<16) + 
			      ((unsigned long) lsc_sg_buf_in[ind][36+7]<<8) +
			      lsc_sg_buf_in[ind][36 + 8] + 10 == data_len)
			cmd_len = 10;
		      else if (((unsigned long) lsc_sg_buf_in[ind][36+4]<<16)+ 
			       ((unsigned long) lsc_sg_buf_in[ind][36+8]<<8)+
			       lsc_sg_buf_in[ind][36 + 9] + 12 == data_len)
			cmd_len = 12;
		      else
			cmd_len = 10;
		    }
		  printf("%s: Assuming command length of %u bytes and reply len %u\n",
			 LSC_LKM_NAME, cmd_len, reply_len);
		}
	      if (data_len < cmd_len)
		{
		  printf("Error data_len (%d) < cmd_len (%d)\n",
			 data_len, cmd_len);
		  error = EINVAL;
		}
	      else
		{
		  data_len -= cmd_len;
		  if (data_len == 0 && reply_len != 0)
		    {
#if 0
		      printf("%s: Data in, data_len == %d, cmd_len = %d, reply_len == %d\n",
			     LSC_LKM_NAME, data_len, cmd_len, reply_len);
#endif
		      error = scsipi_command(lsc_link[ind],
					     (struct scsipi_generic *)
					     (lsc_sg_buf_in[ind] + 36),
					     cmd_len,
					     lsc_sg_buf_out[ind] + 36,
					     reply_len,
					     0, lsc_sg_timeout[ind],
					     NULL,
					     XS_CTL_SILENT | XS_CTL_DATA_IN);
		    }
		  else
		    {
#if 0
		      printf("%s: Data out, data_len == %d, cmd_len = %d, reply_len == %d\n",
			     LSC_LKM_NAME, data_len, cmd_len, reply_len);
#endif
		      error = scsipi_command(lsc_link[ind],
					     (struct scsipi_generic *)
					     (lsc_sg_buf_in[ind] + 36),
					     cmd_len,
					     data_len != 0 ?
					     lsc_sg_buf_in[ind] + 36 +
					     cmd_len :
					     NULL, 
					     data_len,
					     0, lsc_sg_timeout[ind],
					     NULL,
					     XS_CTL_SILENT | XS_CTL_DATA_OUT);
		    }
		}
	    }
	  else
	    {
#if 0
	      printf("%s: Data in2, cmd_len = %d, reply_len == %d\n",
		     LSC_LKM_NAME, cmd_len, reply_len);
#endif
	      error = scsipi_command(lsc_link[ind],
				     (struct scsipi_generic *)
				     (lsc_sg_buf_in[ind] + 36),
				     cmd_len,
				     lsc_sg_buf_out[ind] + 36,
				     reply_len,
				     0, lsc_sg_timeout[ind],
				     NULL,
				     XS_CTL_SILENT | XS_CTL_DATA_IN);
	      *(int *) (lsc_sg_buf_out[ind]) = reply_len + 36;
	      *(int *) (lsc_sg_buf_out[ind] + 4) = reply_len + 36;
	    }
	  *(int *) (lsc_sg_buf_out[ind] + 8) =
	    *(int *) (lsc_sg_buf_in[ind] + 8);
	  *(int *) (lsc_sg_buf_out[ind] + 16) =
	    *(int *) (lsc_sg_buf_in[ind] + 16);
	  if (error)
	    {
#if 0
	      printf("%s: Error %d\n", LSC_LKM_NAME, error);
#endif
	      lsc_sg_buf_len[ind] = 36;
	      *(int *) (lsc_sg_buf_out[ind] + 12) = error; /* Error result */
	      if ((lsc_sg_buf_out[ind][22] & 0xf) != 2 &&
		  (lsc_sg_buf_out[ind][22] & 0xf) != 6)
		{
		  printf("%s: Error %d Sense data: 0x", LSC_LKM_NAME, error);
		  for(i = 0; i < 16; i++)
		    {
		      printf("%02x", lsc_sg_buf_out[ind][20 + i]);
		      if (i % 2 == 1)
			printf(" ");
		    }
		  printf("\n");
		}
	      error = 0;
	    }
	  else
	    {
	      *(int *) (lsc_sg_buf_out[ind] + 12) = 0; /* Result ok */
#ifdef LSC_SG_SCSI_DEBUG
	      if (reply_len != 0)
		{
#ifdef LSC_SG_SCSI_DEBUG_NO_READS
		  if (lsc_sg_buf_in[ind][36] == 0x28)
		    {
		      printf(".");
		    }
		  else
#endif /* LSC_SG_SCSI_DEBUG_NO_READS */
		    {
		      printf("%s: Reply:\n", LSC_LKM_NAME);
		      lsc_hexl_buffer(lsc_sg_buf_out[ind] + 36, reply_len);
		    }
		}
#endif /* LSC_SG_SCSI_DEBUG */
	    }
	}
    }
  else
    {
      error = EIO;
    }
 out:
  splx(s);
  return error;
}

/* This is the select interface that the kernel uses to implement select().
   This checks whether reads/writes are possible (as requested), and
   if so, returns that information immediately.  Otherwise, this records
   that we are in a select, and causes the select to be woken up when
   data is available. */
int lsc_cdevsw_poll(dev_t dev, int rw, struct proc *p)
{
#ifdef LSC_GENERIC_DEBUG
  printf("%s: Poll for device %d, rw = %u\n", LSC_LKM_NAME, minor(dev), rw);
#endif /* LSC_GENERIC_DEBUG */
  return rw & (POLLIN | POLLRDNORM);
}

/* Define a cdevsw structure describing the device. */
#if defined(__NetBSD__)
struct cdevsw lsc_chardev_cdevsw = 
{
        lsc_cdevsw_open,        /* open */
        lsc_cdevsw_close,       /* close */
        lsc_cdevsw_read,        /* read */
        lsc_cdevsw_write,       /* write */
        lsc_cdevsw_ioctl,       /* ioctl */
        (void *)nullop,         /* stop */
        (void *)NULL,           /* tty */
        lsc_cdevsw_poll,        /* poll */
        (void *)nullop,         /* mmap */
        0                       /* type */
};
#elif defined(__FreeBSD__)
struct cdevsw lsc_chardev_cdevsw = {
        lsc_cdevsw_open,
	lsc_cdevsw_close,
	lsc_cdevsw_read,
	lsc_cdevsw_write,
        lsc_cdevsw_ioctl,
	nullstop, noreset,
	nodevtotty,
        lsc_cdevsw_poll,
	nommap,
	nostrategy,
	"lscchardev",
	NULL,
	-1
};
#else /* unknown */
#error Check cdevsw format for this platform
#endif
